#Final Project ChE 210D

#import python modules
import numpy as np
import time
import atomwrite

#import compiled fortran library
import LJG_solvation
#import ljlib

import sys
import os

start = time.clock()

#whether or not to use the 3d visualization
UseVisual = False
#UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 8.5/2.43
#set the random number seed; useful for debugging
#np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    UseVisual = False
	#identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [LJG_solvation.calcenergy(Pos + NormDir * x, M, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = LJG_solvation.calcenergy(Pos + NormDir * dx, M, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [LJG_solvation.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [LJG_solvation.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = LJG_solvation.calcenergy(Pos + NormDir * d3, M, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = LJG_solvation.calcenergy(Pos + NormDir * d3, M, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = LJG_solvation.calcenergyforces(Pos, M, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = LJG_solvation.calcenergyforces(Pos, M, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


#Parameters
epsilon_water = 20.4
sigma_water = 2.43	

N = 216
M = 1
Cut = 8.5/sigma_water

water_density = 1000 #g/L
rho = (water_density*6.02*10**23)/18 #atoms/L
Vol = (N*10**27)/rho #angstroms^3
L = Vol**(1./3.) #angstroms
print L
L = L/sigma_water #scaled to sigma_water
rc = Cut
Dim = 3
print L

#Nondimensionalizing temperature
Temp = 300 # Kelvin
kb = 1.3806503 * 10**-23 # J/K
numerator = (kb * Temp * 6.02 * 10**23)/1000 #Beta
denominator = epsilon_water
beta_nondim = numerator/denominator

coeff = .013 #less than 1
dr = coeff*L
steps = 40100000
outputs = 1000
display_outputs = 1000.
widom_sample_frequency = 250.
equil_period = 100000
number_of_values = 5
#np.random.seed(342324)

#short test
# steps = 2000000
# widom_sample_frequency = 500
# equil_period = 100000

SaveTrajectory = False

def MonteCarlo():
	
	#Initialize positions, minimize cubic lattice energy
	global M, L, N, Cut
	P_acc_count = 0.
	P_acc_Widom = 0.
	exp_part_of_mu_sum = 0.
	count = 0.
	exp_part_of_mu_Array = np.array([])
	i = 0.
	
	if i == 0:
		Pos = InitPositions(N, L)
		print type(Pos)
		#print Pos
		PE, Pos = ConjugateGradient(Pos, .001, 10**-8, 10**-10)
		#print Pos
		print "PE"
		print PE
		print N
		PE_not = PE
		Forces = np.zeros_like(Pos)
		PE, Forces = LJG_solvation.calcenergyforces(Pos, M, L, Cut, Forces)
	
	#Dummy information for video
	Pos_dummy = Pos.copy()
	Pos_dummy = np.append(Pos_dummy,[Pos[len(Pos)-1,:]], axis=0)
	Names = [ 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'C', 'O']
	
	#open file to write to
	data = file('EnergyData''__N=''%s''__M=''%s''__T=''%s''__epsilon_water=''%s''__coeff=''%s''__rho_water=''%s''.txt' % (N,M,Temp,epsilon_water,coeff,water_density), 'w')
	mudata = file('muData''__N=''%s''__M=''%s''__T=''%s''__epsilon_water=''%s''__coeff=''%s''__rho_water=''%s''.txt' % (N,M,Temp,epsilon_water,coeff,water_density), 'w')
	#adata = file('FreeEnergyData_AllRuns.txt', 'a')
	Pdb = atomwrite.pdbfile("widom_insertion.pdb", L, AtomNames = Names)
	Pdb.write(Pos_dummy)
	
	#Monte carlo loop
	while i < steps:
		MoveAtom = np.random.randint(len(Pos))
		OldPos = Pos[MoveAtom,:].copy() #old position of just one atom
		value = 1. * OldPos

		#Calculate energy before
		PE_one_atom_before = LJG_solvation.calcenergyoneposition(Pos, M, L, Cut, MoveAtom)
		#Move random atom
		Pos[MoveAtom,0] += (2*np.random.random()-1)*dr
		Pos[MoveAtom,1] += (2*np.random.random()-1)*dr
		Pos[MoveAtom,2] += (2*np.random.random()-1)*dr

		#Calculate energy after
		PE_one_atom_after = LJG_solvation.calcenergyoneposition(Pos, M, L, Cut, MoveAtom)

		deltaPE = PE_one_atom_after - PE_one_atom_before

		if deltaPE <=0:
			PE = PE + deltaPE
			P_acc_count += 1.
		else:
			P_acc = np.exp(-deltaPE/beta_nondim)#Temp)
			P_Metropolis = np.random.random()
			if P_acc > P_Metropolis:
				PE = PE + deltaPE
				P_acc_count += 1.
			else:
				Pos[MoveAtom,:] = OldPos
		
		#Pdbwriting for video
		# if np.mod(i,400) == 0:
			# Pos_dummy = Pos.copy()
			# Pos_dummy = np.append(Pos_dummy,[Pos[len(Pos)-1,:]], axis=0)
			# Pdb.write(Pos_dummy)
		
		#Widom insertion
		if i >= equil_period and np.mod(i, widom_sample_frequency) == 0:
			count += 1
			#Insert random particle
			Pos = np.append(Pos,[[(2*np.random.random()-1)*L,(2*np.random.random()-1)*L,(2*np.random.random()-1)*L]], axis=0)
			Pos_methane = Pos[len(Pos)-1,:].copy()
			#Calculate deltaU_meth
			deltaU_meth = LJG_solvation.calcenergysolvation(Pos, M, L, Cut, len(Pos)-1)
			#Calculate exp_part_of_mu
			exp_part_of_mu = np.exp(-deltaU_meth/beta_nondim) #.1222 comes from nondimensional T at T = 300K
			exp_part_of_mu_Array = np.append(exp_part_of_mu_Array,[exp_part_of_mu])
			if exp_part_of_mu == 0:
				P_acc_Widom += 1
			if UseVisual:
				if atomvis.Initialized:
					#update the positions
					atomvis.Update(Pos, L)
				else:
					#initialize the visualization window
					atomvis.Init(Pos, L)
			#Write to Pdb
			# if np.mod(i,2000) == 0:
				# Pos_dummy = Pos.copy()
				# Pdb.write(Pos_dummy)
			#Delete methane
			Pos = np.delete(Pos, len(Pos)-1,0)
			#print "%d  deltaU_meth = %11.4f  exp_part_of_mu = %11.8f" % (i, deltaU_meth, exp_part_of_mu)
			#mudata.write('\n%s\t%s\t%s' % (i,i/N, exp_part_of_mu))
		
		if np.mod(i,steps/outputs) == 0:
			PE, Forces = LJG_solvation.calcenergyforces(Pos, M, L, Cut, Forces)
			data.write('\n%s\t%s\t%s' % (i,i/N, PE))
			if np.mod(i,steps/display_outputs) == 0:
				print "\t%d  PE = %11.4f " % (i,PE)
			
		
		#if using visualization, update the display
		if UseVisual:
			if atomvis.Initialized:
				#update the positions
				atomvis.Update(Pos, L)
			else:
				#initialize the visualization window
				atomvis.Init(Pos, L)
		i = i+1

	#print Forces
	print "\nPE_not = %f" % (PE_not)
	print "\nPE_final = %f " % (PE)

	#Calculate acceptance ratio
	P_acc_ratio = P_acc_count/steps
	print "P_acc_ratio"
	print P_acc_ratio
	print "rho"
	print rho
	print "dr"
	print dr
	print "Widom count"
	print P_acc_Widom

	
	#Calculate widom ratio
	print "P_acc_ratio_Widom"
	P_acc_ratio_Widom = 1. - P_acc_Widom/count
	print P_acc_ratio_Widom
	
	#Calculate average of exponential part of Widom insertion
	#exp_part_of_mu_mean = exp_part_of_mu_sum/count
	exp_part_of_mu_mean = exp_part_of_mu_Array.mean(axis=0)
	print "exp_part_of_mu_Array mean"
	print exp_part_of_mu_mean
	
	print "exp_part_of_mu_Array length:"
	print len(exp_part_of_mu_Array)
	
	
	#Calculate averages of exponential part of Widom insertion over a long run split up
	exp_averages_Array = np.array([])
	j = 0
	k = 0
	
	samples = len(exp_part_of_mu_Array)/number_of_values
	while j < number_of_values: #len(exp_part_of_mu_Array)/number_of_values:  
		temp_sum = 0
		counter = j * samples
		while counter < (j+1)*samples:
			temp_sum = temp_sum + exp_part_of_mu_Array[counter]
			counter += 1
		exp_averages_Array = np.append(exp_averages_Array,[temp_sum/samples])
		j += 1
	
	print exp_averages_Array
			
	#Calculate mu_nondims
	mu_nondim_Array = np.array([])
	for ctr in exp_averages_Array:
		mu_nondim_Array = np.append(mu_nondim_Array,[-np.log(ctr)*beta_nondim])
	print "mu_nondim_Array"
	print mu_nondim_Array

	#Calculate mu_dims
	mu_dimensional_Array = epsilon_water * mu_nondim_Array
	print "mu_dimensional_Array"
	print mu_dimensional_Array
	
	#Calculate mu_avg which is the free energy
	mu_avg = mu_dimensional_Array.mean(axis=0)
	
	print "mu_avg = free energy of solvation"
	print mu_avg
	
	#Write mudata to file
	mudata.write('\nepsilon_water = %s\nsigma_water = %s' % (epsilon_water, sigma_water))
	mudata.write('\n%s' % (mu_nondim_Array))
	mudata.write('\n%s' % (mu_dimensional_Array))
	mudata.write('\nmu_avg = free energy of solvation (kJ/mol) = %s' % (mu_avg))
	
	
	data.write('\n%s' % (P_acc_ratio))
	
	data.close()
	mudata.close()
	#adata.close()
	Pdb.close()
	

	
MonteCarlo()		

elapsed = (time.clock() - start)
print elapsed/60
print "minutes"
